@@ -9,8 +9,11 @@
  * Portions Copyright (c) 2003, Atsushi Mitani
  *--------------------------------------------------------------------
  */
+#ifdef USE_REPLICATION
+
 #include "postgres.h"
 
+#include <pthread.h>
 #include <stdio.h>
 #include <string.h>
 #include <unistd.h>
@@ -69,26 +72,41 @@
  * GLOBAL VARIABLE DECLARATION
  *--------------------------------------
  */
+/* for replicate_com.h */
+PGR_Com_Info Com_Info;
+
+/* replication server data */
+uint16_t Port_Number = 0;
+uint16_t LifeCheck_Port_Number = 0;
+uint16_t Recovery_Port_Number = 0;
+
+/* global table data */
 HostTbl *Host_Tbl_Begin = NULL;
 TransactionTbl * Transaction_Tbl_Begin = NULL;
 TransactionTbl * Transaction_Tbl_End = NULL;
+TransactionTbl * PGconn_Tbl_Begin = NULL;
+TransactionTbl * PGconn_Tbl_End = NULL;
 RecoveryTbl * LoadBalanceTbl = NULL;
 RecoveryStatusInf * Recovery_Status_Inf = NULL;
+LockWaitInf * Lock_Wait_Tbl = NULL;
 ReplicateHeader * PGR_Log_Header = NULL;
 ReplicateServerInfo * Cascade_Tbl = NULL;;
 CommitLogInf * Commit_Log_Tbl = NULL;
-ReplicationLogInf * Replicateion_Log = NULL;
-int Port_Number = 0;
+QueryLogType * Query_Log_Top = NULL;
+QueryLogType * Query_Log_End = NULL;
 CascadeInf * Cascade_Inf = NULL;
-int Recovery_Port_Number = 0;
+ReplicationLogInf * Replicateion_Log = NULL;
+/* IPC's id data */
 int RecoveryShmid = 0;
-int RecoveryMsgid = 0;
 int HostTblShmid = 0;
+int LockWaitTblShmid = 0;
+int LoadBalanceTblShmid = 0;
 int CascadeTblShmid = 0;
 int CascadeInfShmid = 0;
 int CommitLogShmid = 0;
-ConfDataType * ConfData_Top = (ConfDataType *)NULL;
-ConfDataType * ConfData_End = (ConfDataType *)NULL;
+int QueryLogMsgid = 0;
+int QueryLogAnsMsgid = 0;
+int PGconnMsgid = 0;
 int MaxBackends = 0;
 char * PGR_Result = NULL;
 int SemID = 0;
@@ -104,15 +122,14 @@ bool PGR_Use_Replication_Log = false;
 bool	PGR_AutoCommit = true;
 unsigned int * PGR_Send_Query_ID = NULL;
 unsigned int PGR_Query_ID = 0;
-int Transaction_Tbl_Num = 0;
-
-int Log_Print = 0;
-int Debug_Print = 0;
-FILE * LogFp = (FILE *)NULL;
-FILE * StatusFp = (FILE *)NULL;
+RecoveryQueueInf RecoveryQueue;
+char * Backend_Socket_Dir = NULL;
+char * PGRuserName = NULL;
+int Reserved_Connections = 1;
+int Idle_Flag = IDLE_MODE;
+bool Exit_Request = false;
 
 extern char *optarg;
-char * PGRuserName = NULL;
 
 /*--------------------------------------
  * PROTOTYPE DECLARATION
@@ -124,11 +141,12 @@ static void replicate_main(void);
 static void quick_exit(SIGNAL_ARGS);
 static void daemonize(void);
 static void write_pid_file(void);
-static void stop_pgreplicate(void);
 static bool is_exist_pid_file(void);
 static void child_wait(SIGNAL_ARGS);
 static void usage(void);
 
+void PGRstop_pgreplicate(int sig);
+
 /*--------------------------------------------------------------------
  * SYMBOL
  *    replicate_loop()
@@ -144,12 +162,12 @@ static void usage(void);
 static int
 replicate_loop(int fd)
 {
-	char * func = "replicate_loop()";
+	char *func = "replicate_loop()";
 	pid_t pgid = 0;
 	pid_t pid = 0;
 	int sock = -1;
 	int rtn = 0;
-	int cnt = 0;
+	bool exist_sys_log = false;
 
 	PGR_Create_Acception(fd,&sock,"",Port_Number);
 	pgid = getpgid(0);
@@ -166,21 +184,12 @@ replicate_loop(int fd)
 		ReplicateHeader  header;
 		char * query = NULL;
 
-		signal(SIGHUP, quick_exit);	
+		signal(SIGHUP, SIG_IGN);	
 		signal(SIGINT, quick_exit);	
 		signal(SIGQUIT, quick_exit);	
 		signal(SIGTERM, quick_exit);	
 		signal(SIGALRM, quick_exit); 
-		signal(SIGPIPE, quick_exit); 
 		setpgid(0,pgid);
-		
-		if (PGRinit_transaction_table() != STATUS_OK)
-		{
-			show_error("transaction table memory allocate failed");
-			PGR_Close_Sock(&sock);
-			exit(1);
-		}
-
 
 		/* child loop */
 		for (;;)
@@ -188,7 +197,12 @@ replicate_loop(int fd)
 			fd_set	  rmask;
 			struct timeval timeout;
 
-			timeout.tv_sec = PGR_RECV_TIMEOUT;
+			Idle_Flag = IDLE_MODE;
+			if (Exit_Request)
+			{
+				quick_exit(SIGINT);
+			}
+			timeout.tv_sec = 1;
 			timeout.tv_usec = 0;
 			
 			if (query != NULL)
@@ -204,37 +218,46 @@ replicate_loop(int fd)
 			rtn = select(sock+1, &rmask, (fd_set *)NULL, (fd_set *)NULL, &timeout);
 			if (rtn && FD_ISSET(sock, &rmask))
 			{
-				query = NULL;
+				Idle_Flag = BUSY_MODE;
+#ifdef PRINT_DEBUG
+				show_debug("%s:replicate_loop selected",func);
+#endif			
 				query = PGRread_packet(sock,&header);
-				if ((query == NULL) || (header.cmdSts == 0))
+				if ((query == NULL) && (header.cmdSys == 0))
 				{
-					cnt ++;
-					if (cnt < 5)
+#ifdef PRINT_DEBUG
+				show_debug("%s: PGRread_packet failed query[%s] cmdSys[%c]",func,query,header.cmdSys);
+#endif			
+					if (exist_sys_log)
 					{
-						/*
-						PGR_Close_Sock(&sock);
-						PGR_Create_Acception(fd,&sock,"",Port_Number);
-						*/
-						continue;
+						show_error("%s:upper cascade closed",func);
+						memset(&header, 0, sizeof(ReplicateHeader));
+						header.cmdSys = CMD_SYS_CALL;
+						header.cmdSts = CMD_STS_QUERY_SUSPEND;
+						header.query_size = htonl(0);
+						PGRsend_rlog_to_local(&header, NULL);
+						exist_sys_log = false;
 					}
 					else
 					{
-						break;
+#ifdef PRINT_DEBUG
+						show_debug("%s:session closed",func);
+#endif			
 					}
-					/*
-					usleep(PGR_RECV_WAIT_MSEC);
-					continue;
 					break;
-					*/
 				}
-				cnt = 0;
 				switch (header.cmdSys)
 				{
 					case CMD_SYS_REPLICATE:
 						status = PGRdo_replicate(sock,&header,query);
+						PGRset_start_life_check();
 						break;
 					case CMD_SYS_LOG:
-						PGRwrite_rlog(&header, query);
+						exist_sys_log = true;
+						PGRsend_rlog_to_local(&header, query);
+						/* set own replicate id by rlog */
+						PGRset_replication_id(header.replicate_id);
+						PGRsend_notice_rlog_done(sock);
 						break;
 					case  CMD_SYS_CASCADE:
 						PGR_Cascade = true;
@@ -262,22 +285,23 @@ replicate_loop(int fd)
 						break;
 				}
 			}
+			if (query != NULL)
+			{
+				free(query);
+				query = NULL;
+			}
 			if (status == LOOP_END)
 			{
 				break;
 			}
 		}
 		PGR_Close_Sock(&sock);
-		if (query != NULL)
-		{
-			free(query);
-			query = NULL;
-		}
-		PGRclear_connections();
+		PGRclear_transactions();
+		PGRclear_connections( DATA_TEMP_USE );
 #ifdef PRINT_DEBUG
 		show_debug("%s:replicate loop exit",func);
 #endif			
-		exit(0);
+		pthread_exit((void*)0);
 	}
 	else
 	{
@@ -289,11 +313,10 @@ replicate_loop(int fd)
 static void
 startup_replication_server(void)
 {
-	ReplicateHeader  header;
 	char hostName[HOSTNAME_MAX_LENGTH];
-	char userName[USERNAME_MAX_LENGTH];
-	char query[256];
 
+	memset(hostName,0,sizeof(hostName));
+	gethostname(hostName,sizeof(hostName)-1);
 	if (PGRuserName == NULL)
 	{
 		PGRuserName = getenv("LOGNAME");
@@ -304,25 +327,7 @@ startup_replication_server(void)
 				PGRuserName = "postgres";
 		}
 	}
-	memset(&header,0,sizeof(ReplicateHeader));
-	memset(query,0,sizeof(query));
-	memset(hostName,0,sizeof(hostName));
-	memset(userName,0,sizeof(userName));
-	gethostname(hostName,sizeof(hostName)-1);
-	strncpy(userName ,PGRuserName,sizeof(userName)-1);
-	snprintf(query,sizeof(query)-1,"SELECT %s(%d,'%s',%d,%d)",
-			PGR_SYSTEM_COMMAND_FUNC,
-			PGR_STARTUP_REPLICATION_SERVER_FUNC_NO,
-			hostName,
-			Port_Number,
-			Recovery_Port_Number);
-	header.cmdSts = CMD_STS_NOTICE;
-	header.query_id = htonl(PGRget_next_query_id());
-	header.query_size = htonl(strlen(query));
-	memcpy(header.from_host,hostName,sizeof(header.from_host));
-	memcpy(header.userName,userName,sizeof(header.userName));
-	strcpy(header.dbName,"template1");
-	PGRreplicate_packet_send( &header, query,-1);
+	PGRnotice_replication_server(hostName,Port_Number,Recovery_Port_Number,LifeCheck_Port_Number,PGRuserName);
 }
 
 /*--------------------------------------------------------------------
@@ -339,9 +344,7 @@ startup_replication_server(void)
 static void
 replicate_main(void)
 {
-#ifdef PRINT_DEBUG
 	char * func = "replicate_main()";
-#endif			
 	int status;
 	int fd = -1;
 	int rtn;
@@ -352,7 +355,7 @@ replicate_main(void)
 		PGRreplicate_exit(0);
 	}
 #ifdef PRINT_DEBUG
-	show_debug("%s %d port bind OK",func,Port_Number);
+	show_debug("%s:replicate main %d port bind OK",func,Port_Number);
 #endif			
 	
 	/* cascade start up notice */
@@ -369,6 +372,11 @@ replicate_main(void)
 		fd_set	  rmask;
 		struct timeval timeout;
 
+		Idle_Flag = IDLE_MODE;
+		if (Exit_Request)
+		{
+			PGRreplicate_exit(SIGINT);
+		}
 		timeout.tv_sec = PGR_RECV_TIMEOUT;
 		timeout.tv_usec = 0;
 
@@ -381,10 +389,15 @@ replicate_main(void)
 		rtn = select(fd+1, &rmask, (fd_set *)NULL, (fd_set *)NULL, &timeout);
 		if (rtn && FD_ISSET(fd, &rmask))
 		{
+			Idle_Flag = BUSY_MODE;
 			/*
 			 * get recovery status.
 			 */
 			PGRcheck_recovered_host();
+			/* 
+			 * get new db login info
+			 */
+			PGRget_pgconn_queue();
 			/*
 			 * call replication module
 			 */
@@ -405,8 +418,16 @@ replicate_main(void)
  *--------------------------------------------------------------------
  */
 static void
-quick_exit(SIGNAL_ARGS)
+quick_exit(int sig)
 {
+	Exit_Request = true;
+	if (sig == SIGTERM)
+	{
+		if (Idle_Flag == BUSY_MODE)
+		{
+			return;
+		}
+	}
 	exit(0);
 }
 
@@ -443,7 +464,7 @@ daemonize(void)
 #ifdef HAVE_SETSID
 	if (setsid() < 0)
 	{
-		show_error("%s:setsid() failed. reason:%s", func,strerror(errno));
+		show_error("%s:setsid() failed. reason:%s",func,strerror(errno));
 		exit(1);
 	}
 #endif
@@ -480,7 +501,7 @@ write_pid_file(void)
 	if (!fd)
 	{
 		show_error("%s:could not open pid file as %s. reason: %s",
-				   func, fname, strerror(errno));
+				   func,fname, strerror(errno));
 		exit(1);
 	}
 	snprintf(pidbuf, sizeof(pidbuf), "%d", getpid());
@@ -495,7 +516,7 @@ write_pid_file(void)
 
 /*--------------------------------------------------------------------
  * SYMBOL
- *    stop_pgreplicate()
+ *    PGRstop_pgreplicate()
  * NOTES
  *    Stop the pgreplicate process
  * ARGS
@@ -504,15 +525,21 @@ write_pid_file(void)
  *    none
  *--------------------------------------------------------------------
  */
-static void 
-stop_pgreplicate(void)
+void 
+PGRstop_pgreplicate(int sig)
 {
-	char * func = "stop_pgreplicate()";
+	char * func = "PGRstop_pgreplicate()";
 	FILE *fd;
 	char fname[256];
 	char pidbuf[128];
 	pid_t pid;
 
+	if (sig == 0)
+	{
+		sig = SIGTERM;
+	}
+	signal(sig,SIG_IGN);
+
 	snprintf(fname, sizeof(fname), "%s/%s", PGR_Write_Path, PGREPLICATE_PID_FILE);
 	fd = fopen(fname, "r");
 	if (!fd)
@@ -526,11 +553,13 @@ stop_pgreplicate(void)
 	fclose(fd);
 	pid = atoi(pidbuf);
 
-	if (kill (pid,SIGTERM) == -1)
+	if (kill (pid,sig) == -1)
 	{
 		show_error("%s:could not stop pid: %d, reason: %s",func,pid,strerror(errno));
 		exit(1);
 	}
+	if (sig != SIGTERM)
+		exit(0);
 }
 
 /*--------------------------------------------------------------------
@@ -607,13 +636,14 @@ usage(void)
 		path = ".";
 	fprintf(stderr,"PGReplicate version [%s]\n",PGREPLICATE_VERSION);
 	fprintf(stderr,"A replication server for cluster DB servers (based on PostgreSQL)\n\n");
-	fprintf(stderr,"usage: pgreplicate [-D path_of_config_file] [-W path_of_work_files] [-U login user][-l][-n][-v][-h][stop]\n");
+	fprintf(stderr,"usage: pgreplicate [-D path_of_config_file] [-W path_of_work_files] [-U login user][-l][-n][-v][-h][stop][restart]\n");
 	fprintf(stderr,"    config file default path: %s/%s\n",path, PGREPLICATE_CONF_FILE);
 	fprintf(stderr,"    -l: print error logs in the log file.\n");
 	fprintf(stderr,"    -n: don't run in daemon mode.\n");
 	fprintf(stderr,"    -v: debug mode. need '-n' flag\n");
 	fprintf(stderr,"    -h: print this help\n");
 	fprintf(stderr,"    stop: stop pgreplicate\n");
+	fprintf(stderr,"    restart: restart pgreplicate\n");
 }
 
 /*--------------------------------------------------------------------
@@ -638,11 +668,18 @@ main(int argc, char * argv[])
 	char * r_path = NULL;
 	char * w_path = NULL;
 	bool detach = true;
+	pid_t syn_pid = 0;
+	pid_t ack_pid = 0;
+	pid_t rlog_pid = 0;
+	int sig_opt = SIGTERM;
+
+	/* init global variables */
+	PGR_Init_Com_Info(&Com_Info);
 
 	r_path = getenv("PGDATA");
 	if (r_path == NULL)
 		r_path = ".";
-	while ((opt = getopt(argc, argv, "U:D:W:lvnh")) != -1)
+	while ((opt = getopt(argc, argv, "U:D:W:m:lvnh")) != -1)
 	{
 		switch (opt)
 		{
@@ -652,7 +689,7 @@ main(int argc, char * argv[])
 					usage();
 					exit(1);
 				}
-				PGRuserName = optarg;
+				PGRuserName = strdup(optarg);
 				break;
 			case 'D':
 				if (!optarg)
@@ -670,11 +707,22 @@ main(int argc, char * argv[])
 				}
 				w_path = optarg;
 				break;
+			case 'm':
+				if (!optarg)
+				{
+					usage();
+					exit(1);
+				}
+				if (!strcmp(optarg,"i"))
+				{
+					sig_opt = SIGINT;
+				}
+				break;
 			case 'l':
-				Log_Print = 1;
+				Com_Info.Log_Info.Log_Print = 1;
 				break;
 			case 'v':
-				Debug_Print = 1;
+				Com_Info.Log_Info.Debug_Print = 1;
 				break;
 			case 'n':
 				detach = false;
@@ -700,9 +748,13 @@ main(int argc, char * argv[])
 
 	if (optind == (argc-1) && !strcmp(argv[optind],"stop"))
 	{
-		stop_pgreplicate();
+		PGRstop_pgreplicate(sig_opt);
 		exit(0);
 	}
+	if (optind == (argc-1) && !strcmp(argv[optind],"restart"))
+	{
+		PGRstop_pgreplicate(SIGTERM);
+	}
 	else if (optind == argc)
 	{
 		if (is_exist_pid_file())
@@ -733,6 +785,37 @@ main(int argc, char * argv[])
 		show_error("%s:PGRinit_recovery error",func);
 		PGRreplicate_exit(0);
 	}
+	if (PGRsyn_init() != STATUS_OK)
+	{
+		show_error("%s:syn_init is failed",func);
+		PGRreplicate_exit(0);
+	}
+
+	/* life check process start */
+	syn_pid = PGR_SYN_Main(SYN_TO_CLUSTER_DB, (void *)Host_Tbl_Begin);
+	if (syn_pid < 0)
+	{
+		show_error("%s:PGR_SYN_Main failed",func);
+		PGRreplicate_exit(0);
+	}
+	ack_pid = PGR_ACK_Main(LifeCheck_Port_Number);
+	if (ack_pid < 0)
+	{
+		show_error("%s:PGR_ACK_Main failed",func);
+		PGRreplicate_exit(0);
+	}
+	if ( PGR_Use_Replication_Log == true )
+	{
+#ifdef PRINT_DEBUG
+		show_debug("Use Replication Log. Start PGR_RLog_Main()");
+#endif
+		rlog_pid = PGR_RLog_Main();
+		if (rlog_pid < 0)
+		{
+			show_error("%s:PGR_RLog_Main failed",func);
+			PGRreplicate_exit(0);
+		}
+	}
 	pgid = getpgid(0);
 	pid = fork();
 	if (pid == (pid_t) -1)
@@ -742,11 +825,11 @@ main(int argc, char * argv[])
 	}
 	if (pid == 0)
 	{
-		signal(SIGHUP, quick_exit);	
+		signal(SIGHUP, SIG_IGN);	
 		signal(SIGINT, quick_exit);	
 		signal(SIGQUIT, quick_exit);	
 		signal(SIGTERM, quick_exit);	
-		signal(SIGPIPE, SIG_IGN);	
+		signal(SIGPIPE, SIG_IGN);
 		/*
 		 * in child process,
 		 * call recovery module
@@ -756,18 +839,22 @@ main(int argc, char * argv[])
 	}
 	else
 	{
+		signal(SIGSEGV, PGRreplicate_exit);
 		signal(SIGINT, PGRreplicate_exit);
 		signal(SIGQUIT, PGRreplicate_exit);
 		signal(SIGTERM, PGRreplicate_exit);
 		signal(SIGCHLD, child_wait);
-		signal(SIGPIPE, SIG_IGN);	
+		signal(SIGPIPE, SIG_IGN);
 		/*
 		 * call replicate module
 		 */
+		Replicateion_Log->r_log_sock = -1;
 		replicate_main();
 	}
 
 	PGRreplicate_exit(0);
 	return STATUS_OK;
 }
-
+#else
+void main() { }
+#endif /* USE_REPLICATION */
